<!DOCTYPE html>
<html>
<head>
  <title>USER SERVER</title>
  <script src="{{STATIC_URL}}jaobi/js/interphone.js" type="text/javascript" charset="UTF-8"></script>
  <script type="text/javascript">
    var server = new interphone({});
  </script>
</head>
<body>
</body>
</html>
